<?php

namespace FiloBlu\ProductUrlTools\Console\Command;

use Exception;
use Magento\Catalog\Model\Product\Attribute\Source\Status;
use Magento\Catalog\Model\Product\Visibility;
use Magento\Catalog\Model\ResourceModel\Product\Action;
use Magento\Catalog\Model\ResourceModel\Product\ActionFactory;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory;
use Magento\CatalogUrlRewrite\Model\ProductUrlRewriteGenerator;
use Magento\Framework\App\ResourceConnection;
use Magento\UrlRewrite\Model\UrlPersistInterface;
use Magento\UrlRewrite\Service\V1\Data\UrlRewrite;
use Psr\Log\LoggerInterface;
use Symfony\Component\Console\Command\Command;
use Symfony\Component\Console\Input\InputArgument;
use Symfony\Component\Console\Input\InputInterface;
use Symfony\Component\Console\Input\InputOption;
use Symfony\Component\Console\Output\OutputInterface;
use Symfony\Component\Console\Helper\ProgressBar;
use Symfony\Component\Console\Style\SymfonyStyle;
use Magento\UrlRewrite\Model\UrlRewriteFactory;
use Symfony\Component\Console\Question\Question;

/**
 * Class RegenerateProductUrlKeyCommand
 * @package FiloBlu\ProductUrlTools\Console\Command
 */
class RegenerateProductUrlKeyCommand extends Command
{
    /**
     * @var CollectionFactory
     */
    protected $collectionFactory;

    /**
     * @var LoggerInterface
     */
    protected $logger;

    /**
     * @var Action
     */
    protected $action;

    /**
     * @var ResourceConnection
     */
    protected $resource;

    /**
     * @var Status
     */
    protected $productStatus;

    /**
     * @var Visibility
     */
    protected $productVisibility;

    /**
     * @var ActionFactory
     */
    protected $actionFactory;

    /**
     * @var ExecutorGroup
     */
    protected $executorGroup;

    /**
     * @var UrlRewriteFactory
     */
    protected $urlRewriteFactory;

    /**
     * @var SymfonyStyle $io
     */
    protected $io;

    /**
     * @var UrlPersistInterface
     */
    protected $urlPersist;

    /**
     * @var ProductUrlRewriteGenerator
     */
    protected $productUrlRewriteGenerator;

    /**
     * @var array
     */
    private $oldUrls;

    /**
     * @var array
     */
    private $urlKeys;

    /**
     * RegenerateProductUrlKeyCommand constructor.
     * @param CollectionFactory $collectionFactory
     * @param LoggerInterface $logger
     * @param Status $productStatus
     * @param Visibility $productVisibility
     * @param ResourceConnection $resource
     * @param ActionFactory $actionFactory
     * @param ExecutorGroup $executorGroup
     * @param UrlRewriteFactory $urlRewriteFactory
     */
    public function __construct(
        CollectionFactory $collectionFactory,
        LoggerInterface $logger,
        Status $productStatus,
        Visibility $productVisibility,
        ResourceConnection $resource,
        ActionFactory $actionFactory,
        ExecutorGroup $executorGroup,
        UrlRewriteFactory $urlRewriteFactory,
        UrlPersistInterface $urlPersist,
        ProductUrlRewriteGenerator $productUrlRewriteGenerator
    )
    {
        $this->collectionFactory = $collectionFactory;
        $this->logger = $logger;
        $this->productVisibility = $productVisibility;
        $this->productStatus = $productStatus;
        $this->actionFactory = $actionFactory;
        $this->resource = $resource;
        $this->executorGroup = $executorGroup;
        $this->urlRewriteFactory = $urlRewriteFactory;
        $this->urlPersist = $urlPersist;
        $this->productUrlRewriteGenerator = $productUrlRewriteGenerator;
        $this->oldUrls = [];
        $this->urlKeys = [];

        parent::__construct();
    }


    /**
     * @param InputInterface $input
     * @param OutputInterface $output
     * @return int|void|null
     */
    public function execute(InputInterface $input, OutputInterface $output)
    {
        $storesList = $this->getStoresToProcess($input->getOption('store'));
        $productIds = $input->getArgument('pids');
        $allVisible = $input->getOption('all_visible');
        $jobs = $input->getOption('jobs');
        $this->action = $this->actionFactory->create();

        if (isset($productIds[0]) && $productIds[0] == 'pids') {
            unset($productIds[0]);
            $productIds = array_values($productIds);
        }

        $helper = $this->getHelper('question');

        $this->io = new SymfonyStyle($input, $output);


        $generate301Redirect = $this->io->ask('Do you want to allow system to generate 301 rediects for the new urls ? (Yes|No)', 'No');
        $generate301RedirectConfig = false;

        $userDefaultRewriteConfig = true;
        $attributeName = '';

        if (strtolower(trim($generate301Redirect[0])) == 'y' && !preg_match("/\W.*/im", $generate301Redirect)) {
            $generate301RedirectConfig = true;

            $useDefaultConfig = $this->io->ask('Do you want to use the default url configuration consisting on product_name-store_id-product_id  ? (Yes|No)', 'Yes');

            if (strtolower(trim($useDefaultConfig[0])) == 'n' && !preg_match("/\W.*/im", $useDefaultConfig)) {
                $userDefaultRewriteConfig = false;

                $this->io->success('You have chosen not to use the default configuration, the variant consists of product_name--selected_attribute_value');

                $attributeExists = true;


                $attributesName = $this->getAttributesName();



                $question = new Question('Please, type the attribute name to use (type @ to exit) : ', 'entity_id');
                $question->setAutocompleterValues($attributesName);

                $attributesNames = [];
                $attributeName = '';
                do {
                    $attributeName = $helper->ask($input, $output, $question);

                    if ($attributeName !== '@') {
                        $attributesNames[] = $attributeName;
                    }

                } while ($attributeName != '@');
            }
        }

        if ($jobs <= 1) {
            $section = null;

            //Rgenerate url Keys
            foreach ($storesList as $storeId) {
                $this->regenerateUrlKey($storeId, $allVisible, $productIds, $section, $userDefaultRewriteConfig, $attributesNames, $generate301RedirectConfig);
            }

            //Regenerate urls
            if (!$userDefaultRewriteConfig) {
                foreach ($storesList as $storeId) {
                    $this->regenerateUrls($storeId, $allVisible, $productIds, $section, $userDefaultRewriteConfig, $attributesNames, $generate301RedirectConfig);
                }
            }

            $output->writeln('<info>✌  Apposc`t!</info>');
            return;
        }

        $jobsArguments = [];

        $i = 0;
        foreach ($storesList as $storeId) {
            $jobsArguments[$i++ % $jobs][] = $storeId;
        }

        foreach ($jobsArguments as $job => $argument) {
            $this->executorGroup->addPhpCommand(sprintf('%s %s %s %s', 'bin/magento', $this->getName(), '-s', implode(',', $argument)));
        }

        $this->executorGroup->executeAndWaitAll();
    }

    /**
     * @param $storesId comma separate store ids
     * @return array
     */
    protected function getStoresToProcess($storesId)
    {
        $allStores = $this->getAllStoreIds();

        if ($storesId === null) {
            return $allStores;
        }

        $stores = explode(',', $storesId);

        $output = [];
        foreach ($stores as $store) {
            $storeId = (int)$store;
            $output[$storeId] = $allStores[$storeId];
        }
        return $output;

    }

    /**
     * @return array
     */
    public function getAllStoreIds()
    {
        $result = [];

        $sql = $this->resource->getConnection()->select()
            ->from($this->resource->getTableName('store'), ['store_id', 'code'])
            ->order('store_id', 'ASC');

        foreach ($this->resource->getConnection()->fetchAll($sql) as $row) {
            if (isset($row['store_id']) && (int)$row['store_id'] > 0) {
                $result[(int)$row['store_id']] = $row['store_id'];
            }
        }

        return $result;
    }

    public function getAttributesName()
    {
        $result = [];

        $sql = $this->resource->getConnection()->select()
            ->from('eav_attribute')
            ->where('backend_type IN ("text", "varchar")');

        foreach ($this->resource->getConnection()->fetchAll($sql) as $row) {
            $result[(int)$row['attribute_id']] = $row['attribute_code'];
        }

        $result[] = "entity_id";

        return $result;
    }

    /**
     * @param $storeId
     * @param $allVisible
     * @param $productIds
     * @param OutputInterface|null $output
     * @param bool $useDefault
     * @param string $attributeName
     * @param $generate301Redirect
     */
    protected function regenerateUrlKey($storeId, $allVisible, $productIds, OutputInterface $output = null, $useDefault = true, $attributesNames = [], $generate301Redirect)
    {
        $collection = $this->collectionFactory->create();

        $collection->setStore($storeId);

        if (!empty($productIds)) {
            $collection->addIdFilter($productIds);
        }

        $collection->addAttributeToSelect(array_merge(['name', 'url_key'], $attributesNames));
        $collection->addStoreFilter($storeId)->setStoreId($storeId);

        if ($allVisible) {
            $collection->addAttributeToFilter('status', ['in' => $this->productStatus->getVisibleStatusIds()]);
            $collection->setVisibility($this->productVisibility->getVisibleInSiteIds());
        }

        $generated = false;


        //$progressBar = new ProgressBar($output, $collection->count());
        //$progressBar->start();

        foreach ($collection->getItems() as $product) {

            $errorFlag = false;

            if (!$product->getName()) {
                $this->logger->info('name is empty for product : ' . $product->getId() . ' in store view ' . $storeId);
                continue;
            }
            $name = $product->getName();


            if ($useDefault) {
                $urlKey = $product->formatUrlKey($name . '-' . $storeId . $product->getId());
            } else {

                $attributes = $product->getData();

                $filteredAttributes = [];

                foreach ($attributesNames as $attributeName) {
                    if (isset($attributes[$attributeName]) && $attributes[$attributeName]) {
                        $filteredAttributes[$attributeName] = $attributes[$attributeName];
                    }

                }


                foreach ($filteredAttributes as $key => &$attributeName) {
                    if ($attributeName == 'entity_id') {
                        $attributeName = $product->getId();
                    }
                }


                $attribute = str_replace(" ", "-", strtolower(implode("-", $filteredAttributes)));
                $urlKey = $product->formatUrlKey($name . '-' . $attribute);

                /*
                if (isset($attributes[$attributeName]) && $attributes[$attributeName]) {

                } else {
                    $urlKey = $product->formatUrlKey($name . '-' . $storeId . $product->getId());
                }*/
            }

            $oldUrl = $product->getData('url_key');

            $this->oldUrls[$storeId][$product->getId()] = $oldUrl;
            $this->urlKeys[$storeId][$product->getId()] = $urlKey;

            $attributesData = ['url_key' => $urlKey];

            try {
                $this->action->updateAttributes([$product->getId()], $attributesData, $storeId);
                $this->logger->info('processed urlkey for product with id : ' . $product->getId() . ' storeID ' . $storeId);
            } catch (Exception $e) {
                $output->writeln('<error>Duplicated url for ' . $product->getId() . ' ' . $e->getMessage() . ' </error>');
            }
        }
    }

    protected function regenerateUrls($storeId, $allVisible, $productIds, OutputInterface $output = null, $useDefault = true, $attributeName = '', $generate301Redirect)
    {
        $collection = $this->collectionFactory->create();

        $collection->setStore($storeId);

        if (!empty($productIds)) {
            unset($productIds['pids']);
            $collection->addIdFilter($productIds);
        }

        $collection->addAttributeToSelect(['name', 'url_key', $attributeName]);
        $collection->addStoreFilter($storeId)->setStoreId($storeId);

        if ($allVisible) {
            $collection->addAttributeToFilter('status', ['in' => $this->productStatus->getVisibleStatusIds()]);
            $collection->setVisibility($this->productVisibility->getVisibleInSiteIds());
        }

        $generated = false;

        foreach ($collection->getItems() as $product) {

            $errorFlag = false;

            //Remove old redirects
            $this->deleteRedirect($storeId, $product->getId());

            //Generate 301 Redirects
            if ($generate301Redirect) {

                //Generate new base urls
                $product->setStoreId($storeId);

                $this->urlPersist->deleteByData([
                    UrlRewrite::ENTITY_ID => $product->getId(),
                    UrlRewrite::ENTITY_TYPE => ProductUrlRewriteGenerator::ENTITY_TYPE,
                    UrlRewrite::REDIRECT_TYPE => 0,
                    UrlRewrite::STORE_ID => $storeId
                ]);

                try {

                    $urls = $this->productUrlRewriteGenerator->generate($product);
                    $this->urlPersist->replace($urls);
                    $this->logger->info('processed url for product with id : ' . $product->getId() . ' storeID ' . $storeId);

                } catch (Exception $e) {
                    $this->io->writeln('<error>Duplicated url for ' . $product->getId() . ' ' . $e->getMessage() . ' </error>');
                    $errorFlag =true;
                }

                $oldUrl = $this->oldUrls[$storeId][$product->getId()];
                $urlKey = $this->urlKeys[$storeId][$product->getId()];

                if ($oldUrl != $urlKey && !$errorFlag) {
                    $generated = true;

                    $this->generate301Redirect($oldUrl, $urlKey, $storeId, $product->getId());
                }
            }
        }
    }

    /**
     *
     */
    protected function configure()
    {
        $this->setName('tools:regenerateurlkey')
            ->setDescription('Regenerate urlkey for given products')
            ->addArgument(
                'pids',
                InputArgument::IS_ARRAY,
                'Products to regenerate'
            )
            ->addOption(
                'store', 's',
                InputOption::VALUE_REQUIRED,
                'Use the specific Store View'
            )
            ->addOption(
                'all_visible', 'a',
                InputOption::VALUE_OPTIONAL,
                'Use for all product visible'
            )->addOption('jobs', 'j',
                InputOption::VALUE_OPTIONAL,
                'Jobs', 1);
    }

    private function generate301Redirect($oldUrl, $newUrl, $storeId, $productId)
    {
        $urlRewriteModel = $this->urlRewriteFactory->create();
        /* set current store id */
        $urlRewriteModel->setStoreId($storeId);
        /* this url is not created by system so set as 0 */
        $urlRewriteModel->setIsSystem(0);
        /* unique identifier - set random unique value to id path */
        $urlRewriteModel->setIdPath(rand(1, 100000));

        /* set redirect type */
        $urlRewriteModel->setRedirectType("301");

        /* set entity type */
        $urlRewriteModel->setEntityType('tools_custom');

        /* set product id */
        $urlRewriteModel->setEntityId($productId);

        /* set actual url path to target path field */
        $urlRewriteModel->setTargetPath($newUrl.".html");
        /* set requested path which you want to create */
        $urlRewriteModel->setRequestPath($oldUrl.".html");
        /* set current store id */

        try {
            if ($urlRewriteModel->save()) {
                $this->io->writeln("<info>Redirect creato correttamente!</info>");
            } else {
                $this->io->writeln("<error>Imossibile creare il redirect</error>");
            }
        } catch (\Exception $e) {
            $this->io->error(['Error with reidrect from: ' . $newUrl, $e->getMessage()]);
        }

    }

    private function deleteRedirect($storeId, $productId)
    {
        $result = [];

        $sql = $this->resource->getConnection()->select()
            ->from('url_rewrite')
            ->where('store_id =' . $storeId)
            ->where('entity_id = ' . $productId)
            ->where('entity_type = "tools_custom"')
            ->where('redirect_type = 301');

        foreach ($this->resource->getConnection()->fetchAll($sql) as $row) {
            $this->resource->getConnection()->delete('url_rewrite', "entity_id = " . $row['entity_id'] . " AND redirect_type = 301 AND store_id = {$storeId}" );
        }

        return $result;
    }
}

